# Copyright 1999-2015. Parallels IP Holdings GmbH. All Rights Reserved.
# -*- coding: utf-8; mode: python -*-
"""A module to perform analysis how the PAM Configuration Set should be changed"""

import os.path

# peer imports 
from support import testset

try: set = set
except NameError: from compat import dummy_set as set

def JumpBeyondLimitsError(atom):
    return ValueError("atom '%s' is likely about a jump beyond configuration chain" % (atom,))
# enddef JumpBeyondLimitsError


def _isLikeSufficient(atom):
    """checks if the atom behaves like a sufficient.
    This includes not only exact match but rather
    some 'inexact' cases"""
    f = atom.control.getFlag('success')
    return f == 'done'
# enddef _isLikeSufficient

def _isLikeRequired(atom):
    """checks if the atom behaves like a required.
    This includes not only exact match but rather
    some 'inexact' cases"""

    # check for typical auth modules errors
    for code in ('perm_denied', 'auth_err', 'cred_insufficient',
                 'user_unknown', 'maxtries', 'acct_expired',
                 'cred_unavail', 'cred_expired', 'cred_err',
                 'authtok_err'):
        f = atom.control.getFlag(code)
        if f in ('bad', 'die'):
            return True
    else:
        return False # really?
# enddef _isLikeRequired

def getTargetPosition(conf, position, steps):
    assert steps > 0
    # since it's _skips_
    steps += 1
    atom = conf[position][0]
    if atom.type == atom.TYPE_CONF_INCLUDE:
        skip = atom.include_size
    else:
        skip = 0
    while steps > 0:
        position += 1 + skip
        if position >= len(conf):
            raise JumpBeyondLimitsError(atom)
        curatom = conf[position][0]
        if curatom.type == curatom.TYPE_CONF_INCLUDE:
            skip = curatom.include_size
        else:
            skip = 0
        steps -= 1
    # endwhile
    return position
# enddef getTargetPosition

def _skipsDeny(conf, position, deny = 'pam_deny.so'):
    """checks if stack execution passes by pam_deny.so upon the module success"""
    atom = conf[position][0]
    skips = atom.control.jump('success')
    if not skips:
        return False
    stoppos = getTargetPosition(conf, position, skips)
    for position in xrange(position + 1, stoppos):
        datom = conf[position][0]
        if datom.type == datom.TYPE_CONF and datom.module == deny:
            return True
    # endfor
    return False
# enddef _skipsDeny
    
class _BaseAnalyzer(object):

    def __init__(self, facility, module):
        self.facility = facility
        self.module = module

    def dropInstalled(self, services, confSet, fullPath = False):
        """for a given list of services returns a list of
        services which don't make use of the module for the facility.
        """
        result = set()
        if fullPath:
            module = self.module
            checkf = lambda (a, f, i): a.module == module
        else:
            module = os.path.basename(self.module)
            checkf = lambda (a, f, i): os.path.basename(a.module) == module
            
        for svc in services:
            if not testset(checkf, confSet.getConf(svc, self.facility)):
                result.add(svc)
        return result
    # enddef dropInstalled

    def dropNonImportant(self, services, confSet, nonImportant, fromEnd = False):
        """for a given list of services drop those which entirely constist
        of modules known as "non-important" in the conf. chain for the facility"""
        return self._selectByCondition(services, confSet, fromEnd,
                                       lambda l: os.path.basename(l[0].module) not in nonImportant)
    # enddef dropNonImportant

    def selectImportant(self, services, confSet, important, fromEnd = False, fullPath = False):
        """for a given list of services select those which do have
        at least one of "important" module in the conf. chain for the facility"""
        if fullPath:
            checkf = lambda l: l[0].module in important
        else:
            checkf = lambda l: os.path.basename(l[0].module) in important
        return self._selectByCondition(services, confSet, fromEnd, checkf)
    # enddef selectImportant

    def selectModules(self, services, confSet, modulesList, fromEnd = False):
        return self._selectByCondition(services, confSet, fromEnd,
                lambda l: os.path.basename(l[0].module) in modulesList)

    def selectServices(self, confSet):
        """Selects services that potentially should be changed"""
        # FIXME: the entire aproach is somewhat sick. I just don't
        # have a better one at the moment...
        shared = []
        interim = []
        shared = confSet.getCommonServiceNames(self.facility)
        for svc in shared:
            if confSet.getIncludes(svc, self.facility):
                interim.append(svc)
        # endfor

        filt = lambda s: confSet.getConf(s, self.facility)
        services = filter(filt, confSet.getServiceNames())
        autonomous = set()

        # sharing is a dict which keys are common [authorization]
        # services names and values a sets of services names which
        # use that particular common service (well, you may recall
        # 'classes of equivalence' term from your high school math
        # course)

        sharing = dict()

        for svc in services:
            if svc in shared:
                continue
            incs = confSet.getAllIncludes(svc, self.facility)
            commonsvc = shared.intersection(incs)
            commonsvc.difference_update(interim)
            if not commonsvc: # empty list
                autonomous.add(svc)
            elif len(commonsvc) == 1: # only a single shared service (don't count interim!) is included
                cs = commonsvc.pop()
                if not sharing.has_key(cs):
                    sharing[cs] = set()
                sharing[cs].add(svc)
            else:
                raise NotImplementedError("service %s includes two or more common services at once: %s" % \
                                          (svc, commonsvc))
        return (autonomous, sharing)
    # enddef

    def _selectByCondition(self, services, confSet, fromEnd, predicate):
        """returns a dict with selected services as keys and
        selected locations as values. used as backend for
        dropNonImportant and selectImportant
        """
        result = dict()

        for svc in services:
            conf = confSet.getConf(svc, self.facility)
            if fromEnd:
                iterf = xrange(len(conf))
            else:
                iterf = xrange(len(conf) - 1, -1, -1)
                
            for pos in iterf:
                location = conf[pos]
                if predicate(location):
                    result[svc] = pos
                    break # to the outer cycle
        # endfor
        return result
    # enddef _selectByCondition

# endclass _BaseAnalyzer

class AuthConfAnalyzer(_BaseAnalyzer):
    """This class represents an analyzer for 'auth' type of PAM facility"""
    
    def __init__(self, module, permit = 'pam_permit.so', deny = 'pam_deny.so'):
        super(AuthConfAnalyzer, self).__init__('auth', module)
        self._permit = permit
        self._deny = deny
    # enddef __init__

    # Have you ever heard about 'fuzzy logic'?
    # Some people may think it's something related to 'estimations',
    # 'unprecise and human-like decisions'. I'd rather think
    # it's more about 'unpredictable results' :-)
    # the following functions exactly fit this definition
    # Beware!..
    def getAddStyle(self, service, position, confSet, involvedServices):
        conf = confSet.getConf(service, self.facility)
        location = conf[position]

        if _isLikeRequired(location[0]):
            hasTail = len(conf) > position + 1
            if not hasTail:
                # unfortunately we should check if a tail exists not only
                # for the current (perhaps, common) service, but also
                # for all services which uses the current one
                for inv in involvedServices:
                    svcconf = confSet.getConf(inv, self.facility)
                    try:
                        svcpos = svcconf.index(location)
                    except ValueError:
                        # emm... someone tries to fool us
                        continue
                    if svcpos + 1 < len(svcconf):
                        hasTail = True
                        break
                    # endif
                # endfor
            # endif
            if hasTail:
                return "requiredTailed"
            else:
                return "required"
        else:
            # A single, sufficient-like change.
            # TODO: IS THAT ALL??
            return "sufficient"
    # enddef getAddStyle

    def getRemoveStyle(self, service, position, confSet, involvedServices):
        conf = confSet.getConf(service, self.facility)
        location = conf[position]
        (theAtom, src, listpos) = location

        if _isLikeRequired(theAtom):
            if position > 0:
                return 'required'
            else:
                return 'requiredSingle'
        else:
            return 'sufficient'
# endclass AuthConfAnalyzer

class PasswordConfAnalyzer(_BaseAnalyzer):
    """This class represents a PAM configuration analyzer for 'password'
    facility"""
    
    def __init__(self, module):
        super(PasswordConfAnalyzer, self).__init__('password', module)

    def getAddStyle(self, service, position, confSet):
        conf = confSet.getConf(service, self.facility)
        theAtom = conf[position][0]
        if _isLikeRequired(theAtom):
            return "required"
        if _isLikeSufficient(theAtom) or _skipsDeny(conf, position):
            return "sufficient"
        raise NotImplementedError("do know how to deal with atom:\n%s" % (theAtom,))
        # not that much :-). but probably in future, I will implement a more elaborate
        # logic. In fact I'm not sure that the current password service insertion logic is
        # correct indeed.
    # enddef getAddStyle

    def getRemoveStyle(self, service, position, confSet, knownPasswordSetters):
        conf = confSet.getConf(service, self.facility)
        theAtom = conf[position][0]
        if _isLikeRequired(theAtom):
            if len(conf) > 1:
                return "required"
            else:
                return "requiredSingle"
        # endif
        return 'optional'
##        if _isLikeSufficient(theAtom) or _skipsDeny(conf, position):
##            if position > 0:
##                prevAtom = conf[position-1][0]
##                if prevAtom.type == prevAtom.TYPE_CONF and \
##                       os.path.basename(prevAtom.module) in knownPasswordSetters and \
##                       _isLikeRequired(prevAtom):
##                    return "replacedSufficient"
##                # endif
##            # endif
##            return "sufficient"
##        # endif
        raise NotImplementedError("do know how to deal with atom:\n%s" % (theAtom,))
    # enddef getRemoveStyle
# endclass PasswordConfAnalyzer
